#include <u.h>
#include <stdio.h>
#include <libdwarf.h>
#include <dwarf.h>
#include "dat.h"

enum {
	NStack	= 512
};

int
dwarfevalconst(Dwarf_Debug dbg, DwarfCompileUnit *cu, uchar *buf, int n, uintptr *val)
{
	int sz;
	int i, p;
	uchar op;
	uchar *ebuf;
	u64int stack[NStack];
	DwarfOpName *opname;

	i = 0;
	p = 0;
	ebuf = buf+n;
	bzero(stack, sizeof stack);

	while(i<n){
		op = buf[i++];
		if(verbose > 2){
			opname = dwarfopname(op);
			fprintf(stderr, "DWARF stack op: '%s'\n",
				opname?opname->name:"<unknown op>");
		}
		switch(op){
		default:
			opname = dwarfopname(op);
			if(debug || verbose)
				fprintf(stderr, "bad stack op: '%s'\n",
				opname?opname->name:"<unknown op>");
			goto BadOp;
			break;

		case DW_OP_dup:
			if(p+1>NStack)
				goto BadStack;
			stack[p+1] = stack[p];
			p++;
			break;

		case DW_OP_minus:
			if(p-1<0)
				goto BadStack;
			stack[p-1] -= stack[p];
			p--;
			break;

		case DW_OP_plus:
			if(p-1<0)
				goto BadStack;
			stack[p-1] += stack[p];
			p--;
			break;

		case DW_OP_plus_uconst:
			stack[p] += (u64int)dwarfgetunsignedleb128(buf+i, ebuf, &sz);
			i += sz;
			break;

		case DW_OP_addr:
			p++;
			stack[p] = dwarfgetaddress(cu, buf+i, ebuf, &sz);
			i += sz;
			break;

		case DW_OP_const1u:
			p++;
			stack[p] = dwarfgetunsigned8(buf+i, ebuf);
			i++;
			break;

		case DW_OP_const1s:
			p++;
			stack[p] = dwarfgetsigned8(buf+i, ebuf);
			i++;
			break;

		case DW_OP_const2u:
			p++;
			stack[p] = dwarfgetunsigned16(buf+i, ebuf);
			i += 2;
			break;

		case DW_OP_const2s:
			p++;
			stack[p] = dwarfgetsigned16(buf+i, ebuf);
			i += 2;
			break;

		case DW_OP_const4u:
			p++;
			stack[p] = dwarfgetunsigned32(buf+i, ebuf);
			i += 4;
			break;

		case DW_OP_const4s:
			p++;
			stack[p] = dwarfgetsigned32(buf+i, ebuf);
			i += 4;
			break;

		case DW_OP_constu:
			p++;
			stack[p] = dwarfgetunsignedleb128(buf+i, ebuf, &sz);
			i += sz;
			break;

		case DW_OP_consts:
			p++;
			stack[p] = dwarfgetsignedleb128(buf+i, ebuf, &sz);
			i += sz;
			break;

		case DW_OP_lit0:
		case DW_OP_lit1:
		case DW_OP_lit2:
		case DW_OP_lit3:
		case DW_OP_lit4:
		case DW_OP_lit5:
		case DW_OP_lit6:
		case DW_OP_lit7:
		case DW_OP_lit8:
		case DW_OP_lit9:
		case DW_OP_lit10:
		case DW_OP_lit11:
		case DW_OP_lit12:
		case DW_OP_lit13:
		case DW_OP_lit14:
		case DW_OP_lit15:
		case DW_OP_lit16:
		case DW_OP_lit17:
		case DW_OP_lit18:
		case DW_OP_lit19:
		case DW_OP_lit20:
		case DW_OP_lit21:
		case DW_OP_lit22:
		case DW_OP_lit23:
		case DW_OP_lit24:
		case DW_OP_lit25:
		case DW_OP_lit26:
		case DW_OP_lit27:
		case DW_OP_lit28:
		case DW_OP_lit29:
		case DW_OP_lit30:
		case DW_OP_lit31:
			p++;
			stack[p] = op - DW_OP_lit0;
			break;
		}
	}
	*val = (u64int)stack[p];
	return 0;

  BadStack:
	return BadStack;

  BadOp:
	return BadOp;
}
