#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <dwarf.h>
#include "cqctbridge.h"

char* cbname[Vnbase] = {
	[Vundef]              = "error!",
	[Vchar]               = "char",
	[Vshort]	      = "short",
	[Vint]		      = "int",
	[Vlong]		      = "long",
	[Vvlong]	      = "long long",
	[Vuchar]	      = "unsigned char",
	[Vushort]	      = "unsigned short",
	[Vuint]		      = "unsigned int",
	[Vulong]	      = "unsigned long",
	[Vuvlong]	      = "unsigned long long",
	[Vfloat]	      = "float",
	[Vdouble]	      = "double",
	[Vlongdouble]	      = "long double",
	[Vvoid]		      = "void",
};

unsigned stdbasetype [DW_bytesize_cnt][9] = {
	[DW_one][0]				 = Vundef,
	[DW_one][DW_ATE_address]		 = Vundef,
	[DW_one][DW_ATE_boolean]		 = Vundef,
	[DW_one][DW_ATE_complex_float]		 = Vundef,
	[DW_one][DW_ATE_float]			 = Vundef,
	[DW_one][DW_ATE_signed]			 = Vchar,
	[DW_one][DW_ATE_signed_char]		 = Vchar,
	[DW_one][DW_ATE_unsigned]		 = Vuchar,
	[DW_one][DW_ATE_unsigned_char]		 = Vuchar,
	[DW_two][0]				 = Vundef,
	[DW_two][DW_ATE_address]		 = Vundef,
	[DW_two][DW_ATE_boolean]		 = Vundef,
	[DW_two][DW_ATE_complex_float]		 = Vundef,
	[DW_two][DW_ATE_float]			 = Vundef,
	[DW_two][DW_ATE_signed]			 = Vshort,
	[DW_two][DW_ATE_signed_char]		 = Vundef,
	[DW_two][DW_ATE_unsigned]		 = Vushort,
	[DW_two][DW_ATE_unsigned_char]		 = Vundef,
	[DW_four][0]				 = Vundef,
	[DW_four][DW_ATE_address]		 = Vuint,
	[DW_four][DW_ATE_boolean]		 = Vundef,
	[DW_four][DW_ATE_complex_float]		 = Vundef,
	[DW_four][DW_ATE_float]			 = Vfloat,
	[DW_four][DW_ATE_signed]		 = Vint,
	[DW_four][DW_ATE_signed_char]		 = Vundef,
	[DW_four][DW_ATE_unsigned]		 = Vuint,
	[DW_four][DW_ATE_unsigned_char]		 = Vundef,
	[DW_eight][0]				 = Vundef,
	[DW_eight][DW_ATE_address]		 = Vundef,
	[DW_eight][DW_ATE_boolean]		 = Vundef,
	[DW_eight][DW_ATE_complex_float]       	 = Vundef,
	[DW_eight][DW_ATE_float]		 = Vundef,
	[DW_eight][DW_ATE_signed]		 = Vvlong,
	[DW_eight][DW_ATE_signed_char]		 = Vundef,
	[DW_eight][DW_ATE_unsigned]		 = Vuvlong,
	[DW_eight][DW_ATE_unsigned_char]       	 = Vundef,
};

Expr*
newexprsrc(Src *src, unsigned kind, Expr *e1, Expr *e2, Expr *e3, Expr *e4)
{
	Expr *e;

	e = malloc(sizeof(Expr));
	if (e == NULL)
	  _exit (3);
	e->kind = kind;
	e->e1 = e1;
	e->e2 = e2;	
	e->e3 = e3;
	e->e4 = e4;

	if(src)
		e->src = *src;

	return e;
}

Expr*
newexpr(unsigned kind, Expr *e1, Expr *e2, Expr *e3, Expr *e4)
{
	return newexprsrc(0, kind, e1, e2, e3, e4);
}

static unsigned basemod[Vnbase][Enbase] = {
	[Vchar][Eunsigned]     = Vuchar,
	[Vchar][Esigned]       = Vchar,

	[Vshort][Eunsigned]    = Vushort,
	[Vshort][Esigned]      = Vshort,
	[Vshort][Eint]         = Vshort,
	[Vushort][Eint]        = Vushort,

	[Vint][Eunsigned]      = Vuint,
	[Vint][Esigned]        = Vint,
	[Vint][Elong]          = Vlong,
	[Vuint][Elong]         = Vulong,
	[Vint][Eshort]         = Vshort,
	[Vuint][Eshort]        = Vushort,
	
	[Vlong][Eunsigned]     = Vulong,
	[Vlong][Esigned]       = Vlong,
	[Vlong][Elong]         = Vvlong,
	[Vulong][Elong]        = Vuvlong,

	[Vvlong][Eunsigned]    = Vuvlong,
	[Vvlong][Esigned]      = Vvlong,

	[Vdouble][Elong]       = Vlongdouble,

	[Vundef][Echar]        = Vchar,
	[Vundef][Edouble]      = Vdouble,
	[Vundef][Efloat]       = Vfloat,
	[Vundef][Eint]         = Vint,
	[Vundef][Elong]        = Vlong,
	[Vundef][Eshort]       = Vshort,
	[Vundef][Esigned]      = Vint,
	[Vundef][Eunsigned]    = Vuint,
	[Vundef][Evoid]        = Vvoid,
	/* the rest are Vundef, which we assume to be 0 */
};

Cbase
baselist(Expr *e)
{
	Expr *s;
	Cbase base;
	
	base = Vundef;
	while(e->kind != Enull){
		s = e;
		e = e->e1;
		switch(s->kind){
		case Echar:
		case Edouble:
		case Efloat:
		case Eint:	
		case Elong:
		case Eshort:
		case Esigned:
		case Eunsigned:
		case Evoid:
			base = basemod[base][s->kind];
			/* FIXME: can we rely on parser to structure these
			   constructions, and eliminate Vundef? */
			if(base == Vundef) {
			  printf ("bad type specifier");
			  _exit (2);
			}
			break;
		default:
		  printf ("bad type specifier");
		  _exit (1);
		}
	}
	return base;
}

Expr * isstdbasetype (char *);
Expr * stdbasepart (char *);

enum DW_bytesizes
mapsizes (int bsize)
{
  switch (bsize) {
  case 1: return DW_one;
    break;
  case 2: return DW_two;
    break;
  case 4: return DW_four;
    break;
  case 8: return DW_eight;
    break;
  default:
    printf ("Can't handle byte size of %d for base type \n", bsize);
    return DW_bytesize_cnt;
  }
}

struct slist {
  char *s;
  struct slist *prev;
  struct slist *next;
};

typedef struct slist slist;

Expr *
isstdbasetype (char *ins)
{
  char *token, *tend = " ", *s;
  slist *start, *last;
  Expr *e, *rv;
  s = strdup (ins);
  token = strtok (s, tend);
  if (token == NULL)
    return 0;
  start = (slist *) malloc (sizeof (slist));
  if (start == NULL)
    _exit (4);
  memset ((void *) start, 0, sizeof (slist));
  last = start;
  start->s = token;
  while (1) {
    token = strtok (NULL, tend);
    if (token == NULL)
      break;
    last->next = (slist *) malloc (sizeof (slist));
    if (start == NULL)
      _exit (4);
    memset ((void *) last->next, 0, sizeof (slist));
    last->next->prev = last;
    last->next->s = token;
    last = last->next;
  }
  rv = e = stdbasepart (last->s);
  if (e == NULL)
    return 0;
  while (start != last) {
    e->e1 = stdbasepart (start->s);
    if (e->e1 == NULL)
      return 0;
    else {
      e = e->e1;
      start = start->next;
    }
  }
  e->e1 = newexpr (Enull, 0, 0, 0, 0);
  return rv;
}

Expr *
stdbasepart (char *s)
{
  if (!strcmp (s, "char"))
    return newexpr (Echar, 0, 0, 0, 0);
  if (!strcmp (s, "int"))
    return newexpr (Eint, 0, 0, 0, 0);
  if (!strcmp (s, "long"))
    return newexpr (Elong, 0, 0, 0, 0);
  if (!strcmp (s, "short"))
    return newexpr (Eshort, 0, 0, 0, 0);
  if (!strcmp (s, "unsigned"))
    return newexpr (Eunsigned, 0, 0, 0, 0);
  if (!strcmp (s, "signed"))
    return newexpr (Esigned, 0, 0, 0, 0);
  if (!strcmp (s, "float"))
    return newexpr (Efloat, 0, 0, 0, 0);
  if (!strcmp (s, "double"))
    return newexpr (Edouble, 0, 0, 0, 0);
  if (!strcmp (s, "void"))
    return newexpr (Evoid, 0, 0, 0, 0);
  else
    return 0;
}

char *
get_cbname (unsigned int cb)
{
  if (cb <= Vnbase)
    return cbname[(enum Cbase)cb];
  return cbname [(enum Cbase) 0];
}

