#include <u.h>
#include <stdio.h>
#include <dwarf.h>
#include <libdwarf.h>

#include "queue.h"
#include "dat.h"
#include "hash.h"
#include "type.h"

DwarfStack*
mkdwarfstack(void)
{
	DwarfStack *ds;

	ds = malloc(sizeof *ds);
	memset(ds, 0, sizeof *ds);
	ds->nalloc = 1024;
	ds->stk = malloc(ds->nalloc*sizeof(ds->stk[0]));
	memset(ds->stk, 0, ds->nalloc*sizeof(ds->stk[0]));
	return ds;
}

void
dwarffreestack(DwarfStack *ds)
{
	memset(ds->stk, 0xdd, ds->nalloc*sizeof(ds->stk[0]));
	free(ds->stk);
	memset(ds, 0xdd, sizeof *ds);
	free(ds);
	return;
}


void
dwarfpush(DwarfStack *ds)
{
	int n;

	ds->depth++;
	if(ds->depth == ds->nalloc){
		n = ds->nalloc+1024;
		ds->stk = realloc(ds->stk, n*sizeof(ds->stk[0]));
		memset(ds->stk+ds->nalloc, 0, 1024*sizeof(ds->stk[0]));
		ds->nalloc = n;
	}
	return;
}


void
dwarfpop(DwarfStack *ds)
{
	if(ds->depth == 0)
		abort();
	ds->depth--;
}


int
dwarfdepth(DwarfStack *ds)
{
	return ds->depth;
}


void
dwarfsettop(DwarfStack *ds, DwarfType *dt)
{
	ds->stk[ds->depth] = dt;
	return;
}


void
dwarfdumpstack(DwarfStack *ds)
{
	USED(ds);
	return;
}

DwarfType*
dwarffindparent(DwarfStack *ds, int type)
{
	int i;

	for(i=dwarfdepth(ds)-1; i>=0; i--){
		if(ds->stk[i] == nil || ds->stk[i]->type == nil)
			continue;
		if(ds->stk[i]->type->type == type)
			return ds->stk[i];
	}
	return nil;
}

DwarfType*
dwarffindsueparent(DwarfStack *ds)
{
	int i;

	for(i=dwarfdepth(ds)-1; i>=0; i--){
		if(ds->stk[i] == nil || ds->stk[i]->type == nil)
			continue;
		switch(ds->stk[i]->type->type){
		case Tstruct:
		case Tunion:
		case Tenum:
			return ds->stk[i];
		}
	}

	return nil;
}
