#include <u.h>
#include <stdio.h>
#include <dwarf.h>
#include <libdwarf.h>

#include "queue.h"
#include "dat.h"
#include "hash.h"
#include "type.h"
#include "util.h"

DwarfVarList	dwarfvars;

static	int	dwarfresolvesym(DwarfVar*);

void
dwarfvarinit(void)
{
	TAILQ_INIT(&dwarfvars);
	return;
}

DwarfVar*
mkdwarfvar(Decl *d, u64int addr)
{
	DwarfVar *v;

	v = malloc(sizeof *v);
	memset(v, 0, sizeof *v);
	v->addr = addr;
	v->decl = d;
	TAILQ_INSERT_TAIL(&dwarfvars, v, link);
	return v;
}

void
dwarfprintsyms(void)
{
	DwarfVar *v;
	HT *ht;

	ht = mkht();
	TAILQ_FOREACH(v, &dwarfvars, link)
		if(v->addr != 0 && v->decl != nil && v->decl->name
		   && !wasseen(ht, v->decl->name))
			printsym(v);
	freeseen(ht);
	freeht(ht);
}

int
dwarfresolvesyms(void)
{
	int rv;
	DwarfVar *v;

	TAILQ_FOREACH(v, &dwarfvars, link)
		if(v->decl != nil){
			rv = dwarfresolvesym(v);
			if(rv < 0)
				return -1;
		}
	return 0;
}


static int
dwarfresolvesym(DwarfVar *v)
{
	int rv;

	rv = dwarfresolvedecl(v->decl);
	return rv;
}
